/////////////////////////////////////////////////////////////////////////////
//
//   verf.c                                          
//   (DWL)
//
//      Miscelaneous functions to support the "Bejing Demo". SHA-1 is 
//   implemented as 1 function in HashMessage and as Several in InitHash,
//   HashBlob and LastBlob. Also several public key functions are provided
//   for software verification of Digital Signatures and Public Key 
//   Certificates.
//
//   Version 1.00 : Initial creation
//   5/12/97
//
/////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>

#include "verf.h"
#include "bnt.h"

//
// Really big for testing purposes
//
#define MAX_BLOB_SIZE 0x4000
puchar pMsgBuff;
puchar pIndex;
ulong  Sum;

static BOOL HashBuffFree = TRUE;

//
// Application id
//
const char AppId[] = "Beijing Project Version 1.0";
//
// Globals for key set components
//
Big DalPubExp, DalMod;
Big CiBPubExp, CiBMod;

//
// ROM id of button containing CiBPubExp and CiBMod
//
uchar CiBROM[8];

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Stores the public exponent and the modulus of Dallas Semiconductor's
//  RSA key set in global variables. In this case Dallas Semiconductor is 
//  acting as a CA (Certification Authority) to validate public keys.
//
// Parameters:
//     ExpLen : Length of public exponent
//     pExp   : Pointer to public exponent
//     ModLen : Length of modulus
//     pMod   : Pointer to modulus
//
// Return value:
//     none
//
void DLLEXPORT SetDallasPublicKey(ushort ExpLen, puchar pExp, 
                                  ushort ModLen, puchar pMod)
{
   // Dump exponent data in Big number structure
   DalPubExp.Len = ExpLen;
   memcpy(DalPubExp.Digit, pExp, ExpLen);
   // Dump modulus data in Big number structure
   DalMod.Len = ModLen;
   memcpy(DalMod.Digit, pMod, ModLen);
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Retrieves the public exponent and the modulus of Dallas 
//  Semiconductor's RSA key set. 
//
// Parameters:
//     pExpLen : Pointer to length of public exponent
//     pExp    : Pointer to public exponent
//     pModLen : Pointer to length of modulus
//     pMod    : Pointer to modulus
//
// Return value:
//     none
//
void DLLEXPORT GetDallasPublicKey(pushort pExpLen, puchar pExp,
                                  pushort pModLen, puchar pMod)
{
   // Retrieve data from Dallas public exponent
   *pExpLen = DalPubExp.Len;
   memcpy(pExp, DalPubExp.Digit, DalPubExp.Len);
   // Retrieve data from Dallas modulus
   *pModLen = DalMod.Len;
   memcpy(pMod, DalMod.Digit, DalMod.Len);
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Stores the public exponent and the modulus of an as of yet 
//  unvalidated RSA key set in global variables.
//
// Parameters:
//     ExpLen    : Length of public exponent
//     pExp      : Pointer to public exponent
//     ModLen    : Length of modulus
//     pMod      : Pointer to modulus
//     pSerialId : Pointer to the ROM id of the CiB containing this key set
//
// Return value:
//     none
//
void DLLEXPORT SetCiBPublicKey(ushort ExpLen, puchar pExp, ushort ModLen, 
                               puchar pMod, puchar pSerialId)
{
   // Dump exponent data in Big number structure
   CiBPubExp.Len = ExpLen;
   memcpy(CiBPubExp.Digit, pExp, ExpLen);
   // Dump modulus data in Big number structure
   CiBMod.Len = ModLen;
   memcpy(CiBMod.Digit, pMod, ModLen);
   // Copy ROM data
   memcpy(CiBROM, pSerialId, 8);
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Retrieves the public exponent and the modulus of a CiB's RSA key set.
//
// Parameters:
//     pExpLen : Pointer to length of public exponent
//     pExp    : Pointer to public exponent
//     pModLen : Pointer to length of modulus
//     pMod    : Pointer to modulus
//     pSerialId : Pointer to ROM id of CiB containing this key set
//
// Return value:
//     none
//
void DLLEXPORT GetCiBPublicKey(pushort pExpLen, puchar pExp, pushort pModLen,
                               puchar pMod, puchar pSerialId)
{
   // Retrieve data from Dallas public exponent
   *pExpLen = CiBPubExp.Len;
   memcpy(pExp, CiBPubExp.Digit, CiBPubExp.Len);
   // Retrieve data from Dallas modulus
   *pModLen = CiBMod.Len;
   memcpy(pMod, CiBMod.Digit, CiBMod.Len);
   // Retrieve CiB's serial number
   memcpy(pSerialId, CiBROM, 8);
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Attempt to verify the public key contained in CiBPubExp and CiBMod.
//  the certificate (passed in Cert) was generated by signing the hash of 
//  the concatenation of the following data elements:
//
//        Message = Public Exponent + Modulus + ROM id + Application id
//        Certificate = H(Message) ^ DalPrivExp Mod DalMod
//
//  Verification will be performed in two steps. First the certificate is
//  decrypted using the Dallas Semiconductor public key (DalPubExp, DalMod).
//  Next we compute (hopefully) the same hash value, H'(Message), using our
//  copy of the data elements shown above. If the hash value we compute (H')
//  is identical to the hash value contained within the decrypted certificate
//  the public key is assumed to be safe to use for signature verification
//  for the application specified by the Application id (AppId).
//
// Parameters:
//     CertLen : Length of certificate data
//     pCert   : Pointer to certificate data
//    
// Return value:
//     TRUE if the public key verification was successful, FALSE otherwise.
//
BOOL DLLEXPORT VerifyCiBPublicKey(ushort CertLen, puchar pCert)
{
   uchar Message[MSG_LEN_MAX];
   ulong MsgLen;
   Big   PubKeyCert, Plain;
   BOOL  MsgDigestOK = FALSE;
   Hash  MsgDigest;

   // Copy certificate length, data to big number PubKeyCert
   PubKeyCert.Len = CertLen;
   memcpy(PubKeyCert.Digit, pCert, CertLen);
   // Decrypt certificate
   ModExp(PubKeyCert, DalPubExp, DalMod, &Plain);

   // Build message array with type and length specifiers for each field
   MsgLen = 0;
   // Start with the public exponent
   Message[MsgLen++] = CIB_EXPONENT;
   Message[MsgLen++] = (uchar) CiBPubExp.Len;
   memcpy(&Message[MsgLen], CiBPubExp.Digit, CiBPubExp.Len);
   MsgLen += CiBPubExp.Len;
   // Concatenate the modulus
   Message[MsgLen++] = CIB_MODULUS;
   Message[MsgLen++] = (uchar) CiBMod.Len;
   memcpy(&Message[MsgLen], CiBMod.Digit, CiBMod.Len);
   MsgLen += CiBMod.Len;
   // Concatenate the ROM id
   Message[MsgLen++] = CIB_ROM_DATA;
   Message[MsgLen++] = 8;
   memcpy(&Message[MsgLen], CiBROM, 8);
   MsgLen += 8;
   // Finally concatenate the application id
   Message[MsgLen++] = CIB_CONFIG_DATA;
   Message[MsgLen++] = strlen(AppId);
   memcpy(&Message[MsgLen], AppId, strlen(AppId));
   MsgLen += strlen(AppId);

   // Hash the whole pile off goo
   HashMessage(Message, MsgLen, MsgDigest);

   //
   //    What should be returned in the big number Plain has the following
   // structure.
   //
   //          Plain.Digit[0] : Type byte of hash data
   //          Plain.Digit[1] : Length (L) byte of hash data
   //          Plain.Digit[2] -> Plain.Digit[2 + L] : Hash (message digest)
   //        
   // and other stuff we don't care about here.
   //

   // Compare the hash values if hash lengths are identical
   if (Plain.Digit[1] == HASH_LEN)
      MsgDigestOK = !memcmp(&Plain.Digit[2], (puchar) MsgDigest, HASH_LEN);

   return MsgDigestOK;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Attempt to verify the signature pointed to by pSig. The public key
//  used for verification is contained in (CiBPubExp, CiBMod) and is assumed 
//  to have been previously verified by a call to VerifyCertificate. The 
//  signature should contain a message digest (pointed to by pDigest) signed
//  using the CiB's private exponent. If the decrypted contents of the 
//  signature contains a message digest (hash) identical to the digest 
//  pointed to by pDigest the signature is assumed to be valid.
//
// Parameters:
//     SigLen    : Length of the signature 
//     pSig      : Pointer to the signature 
//     pDigest   : Pointer to the message digest
//
// Return value:
//     TRUE if the signature verification was successful, FALSE otherwise.
//
BOOL DLLEXPORT VerifySignature(ushort SigLen, puchar pSig, puchar pDigest)
{
   Big  DigitalSig, Plain;
   BOOL MsgDigestOK = FALSE;
   int  i;

   // Copy signature length, data to big number DigitalSig
   DigitalSig.Len = SigLen;
   memcpy(DigitalSig.Digit, pSig, SigLen);
   // Decrypt signature
   ModExp(DigitalSig, CiBPubExp, CiBMod, &Plain);

   //
   //    What should be returned in the big number Plain has the following
   // structure.
   //
   //          Plain.Digit[0] -> Plain.Digit[L-1] : Hash (message digest)
   //        
   // and other stuff we don't care about here.
   //

   // Reverse byte order within words 
   for (i = 0; i < 20; i += 4)
   {
     Swap(&Plain.Digit[i],   &Plain.Digit[i+3]);
	  Swap(&Plain.Digit[i+1], &Plain.Digit[i+2]);
   }

   // Make sure lengths of digests are the same
   MsgDigestOK = !memcmp(Plain.Digit, pDigest, 20);

   return MsgDigestOK;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Sets the initial hash value and allocates lots of memory to hold
//  caller's message.
//
// Parameters:
//     H : Hash value
//
// Return value:
//     TRUE if piecewise hash not in use by another app, FALSE otherwise.
//
BOOL DLLEXPORT InitHash(Hash H)
{
   if (HashBuffFree)
   {
      // Initialize hash value
      H[0] = 0x67452301;
      H[1] = 0xEFCDAB89;
      H[2] = 0x98BADCFE;
      H[3] = 0x10325476;
      H[4] = 0xC3D2E1F0;

      pMsgBuff = (puchar) calloc(MAX_BLOB_SIZE, sizeof(uchar));

      if (pMsgBuff)
          HashBuffFree = FALSE;

      pIndex = pMsgBuff;
      Sum = 0;
   }

   return !HashBuffFree;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Adds a chunk of a message to the caller's message buffer.
//
// Parameters:
//     pBlob   : pointer to chunk of message data
//     BlobLen : Length of message data chunk
//
// Return value:
//     TRUE if blob can be successfully added to the message buffer.
//
BOOL DLLEXPORT AddToBlob(puchar pBlob, ulong BlobLen)
{
   Sum += BlobLen;

   if (BlobLen >= MAX_BLOB_SIZE)
      return FALSE;

   // Add blob to message buffer
   memcpy(pIndex, pBlob, BlobLen);
   // Move message buffer pointer
   pIndex += BlobLen;

   return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Adds the last chunk of a message to the caller's message buffer and
//  hashes the whole thing.
//
// Parameters:
//     pBlob   : pointer to chunk of message data
//     BlobLen : Length of message data chunk
//     H       : Hash value
//
// Return value:
//     TRUE if blob can be successfully added to the message buffer.
//
BOOL DLLEXPORT HashBlob(puchar pBlob, ulong BlobLen, Hash H)
{
   ulong BlockNum, i;

   BOOL RetVal = FALSE;
   if (AddToBlob(pBlob, BlobLen))
   {
      // Reverse byte order within word 
      for (i = 0; i < Sum; i += 4)
      {
         Swap(&pMsgBuff[i],   &pMsgBuff[i+3]);
	      Swap(&pMsgBuff[i+1], &pMsgBuff[i+2]);
      }

      PadMessage(pMsgBuff, &BlockNum, Sum);

      for (i = 0; i < BlockNum; i++)
         HashBlock((pulong) &pMsgBuff[i*BLOCK_SIZE_B], H);

      RetVal = TRUE;
   }

   HashBuffFree = TRUE;
   free(pMsgBuff);

   return RetVal;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Performs the SHA-1 hash on the message pointed to by pMsgStr.
//
// Parameters:
//     pMsgStr : pointer to message data
//     MsgLen  : Length of message data
//     H       : Hash value
//
// Return value:
//     TRUE if we are able to allocate sufficient memory for the message
//  plus a little extra room for the padding.
//
BOOL DLLEXPORT HashMessage(puchar pMsgStr, ulong MsgLen, Hash H)
{
   puchar PdMsg;
   ulong  BlockNum, i;

   // Initialize hash value
   H[0] = 0x67452301;
   H[1] = 0xEFCDAB89;
   H[2] = 0x98BADCFE;
   H[3] = 0x10325476;
   H[4] = 0xC3D2E1F0;
 
   // Allocate a buffer that is big enough
   PdMsg = (puchar) malloc(MsgLen + 2*BLOCK_SIZE_B);

   if (!PdMsg)
      return FALSE;

   memset(PdMsg, 0, MsgLen + 2*BLOCK_SIZE_B);
	                      
   // Copy original message 
   memcpy(PdMsg, pMsgStr, MsgLen);
   // Reverse byte order within word 
   for (i = 0; i < MsgLen; i += 4)
   {
     Swap(&PdMsg[i],   &PdMsg[i+3]);
	  Swap(&PdMsg[i+1], &PdMsg[i+2]);
   }

   PadMessage(PdMsg, &BlockNum, MsgLen);

   for (i = 0; i < BlockNum; i++)
      HashBlock((pulong) &PdMsg[i*BLOCK_SIZE_B], H);

   free(PdMsg);

   return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     When implemented should lock down the memory buffer used for 
//  piecewise hashing.
//
// Parameters:
//     none
//
// Return value:
//     none
//
void DLLEXPORT LockHash()
{


}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     When implemented should unlock the memory buffer used for 
//  piecewise hashing.
//
// Parameters:
//     none
//
// Return value:
//     none
//
void DLLEXPORT UnlockHash()
{


}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Performs a left circular shift of x by n bits
//
// Parameters:
//     x : value to be shifted
//     n : number of bits in shift
//
// Return value:
//     x <<< n    
//
ulong S(ulong x, uchar n)
{
   return (x << n) | (x >> (32 - n));
}

/////////////////////////////////////////////////////////////////////////////
ulong Ft(int t, ulong B, ulong C, ulong D)
{
   ulong f;

   if (t < 20)
   {
      f = (B & C) | ((~B) & D);
   }
   else if ((t < 40) || (t > 59))
   {
      f = B ^ C ^ D;
   }
   else if (t < 60)
   {
	   f = (B & C) | (B & D) | (C & D);
   }

   return f;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//     Hashes 1 64-byte block of the message.
//
// Parameters:
//    pMB : Pointer to the message block
//    H   : Cummulative hash value
//
// Return value:
//    none
//
void HashBlock(pulong pMB, Hash H)
{
   ulong A, B, C, D, E;
   ulong W[80], Kt, Temp;
   int   t;
   
   memcpy((uchar *) W, (uchar *) pMB, 64);

   for (t = 16; t < 80; t++)
   {
      W[t] = S(W[t-3] ^ W[t-8] ^ W[t-14] ^ W[t-16], 1);         
   }

   A = H[0]; B = H[1]; C = H[2]; D = H[3]; E = H[4];

   for (t = 0; t < 80; t++)
   {
	  if (t < 20)
		  Kt = K0;
	  else if (t < 40)
		  Kt = K20;
	  else if (t < 60)
		  Kt = K40;
	  else
		  Kt = K60;

     // Temp = S(A, 5) + Ft(t, B, C, D) + E + W[t] + Kt;


	  Temp = S(A, 5);
	  Temp += Ft(t, B, C, D);
	  Temp += E;
	  Temp += W[t];
	  Temp += Kt;

	  E = D; D = C;
	  C = S(B, 30);
	  B = A; A = Temp;
   }

   H[0] += A; H[1] += B; H[2] += C; H[3] += D; H[4] += E;
}

/////////////////////////////////////////////////////////////////////////////
//
// Routine description:
//    Pad the message as specified in FIPS 180-1. This procedure assumes 
// that the buffer pointed to by pMsg is large enough to hold up to 64 more 
// bytes.
//
// Parameters:
//    pMsg      : Pointer to unpadded message
//    BlockNum  : Pointer to number of block in message
//    MsgLen    : Length of unpadded message
//
// Return value:
//    none
//
void PadMessage(puchar pMsg, ulong *BlockNum, ulong MsgLen)
{
   pulong        pW;
   uchar         LeftOver;

   *BlockNum = (MsgLen >> 6) + 1;

   LeftOver = (uchar) (64 - MsgLen % 64);

   pMsg[(MsgLen+3) - ((MsgLen%4)*2)] = 0x80;

   if (LeftOver > 8)
   {
	   pW = (pulong) &pMsg[(*BlockNum - 1) << 6];
   }
   else
   {
      pW = (pulong) &pMsg[*BlockNum << 6];
	   ++(*BlockNum);
   }

   // Assign bit length of original message
   pW[14] = 0x00;
   pW[15] = MsgLen << 3;
}

/////////////////////////////////////////////////////////////////////////////
void Swap(uchar *a, uchar *b)
{
   uchar Tmp;

   Tmp = *a;
   *a  = *b;
   *b  = Tmp;
}


